unit mathPanl;
{ Mathematical Panel }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

type
  TxgrType = ( xgrNone, xgrCentered, xgr1quadrant, xgrFlatY, xgrFlatX );
  TxgrPlotMode = ( xgrDotMode, xgrDotModeFast, xgrBarMode );

  TmathPanel = class(Tpanel)
  private
    // ***************************************************************
    //          PRIVATE VARIABLES
    // ***************************************************************
    LastXpixDrawn : integer;
    LastYpixDrawn : integer;
    // Dati sul proporzionamento fisico <-> numerico del grafico
    Xqnt : double;
    Yqnt : double;
    R2Width : integer;
    R2Height : integer;
    //
    fBackColor : Tcolor;
    fAxesColor : Tcolor;
    //
    fMarginOriz : integer;
    fMarginVert : integer;
    fXMin : double;
    fXMax : double;
    fYMin : double;
    fYMax : double;
    //
    fScaleX : boolean;
    fScaleXFont : Tfont;
    fScaleXTags : integer;
    fScaleY : boolean;
    fScaleYFont : Tfont;
    fScaleYTags : integer;
    //
    fGraphType : TxgrType;
    //
    fPlotMode : TxgrPlotMode;
    // ***************************************************************
    //          ACCESS METHODS
    // ***************************************************************
    Procedure SetBackColor( value : Tcolor );
    Procedure SetAxesColor( value : Tcolor );
    //
    Procedure SetMarginOriz( value : integer );
    Procedure SetMarginVert( value : integer );
    //
    Procedure SetXMin( value : double );
    Procedure SetXMax( value : double );
    Procedure SetYMin( value : double );
    Procedure SetYMax( value : double );
    //
    Procedure SetScaleX( value : boolean );
    Procedure SetScaleXFont( value : Tfont );
    Procedure SetScaleXTags( value : integer );
    Procedure SetScaleY( value : boolean );
    Procedure SetScaleYFont( value : Tfont );
    Procedure SetScaleYTags( value : integer );
    //
    Procedure SetGraphType( value : TxgrType );
    //
    Procedure SetPlotMode( value : TxgrPlotMode );
    //
    Procedure AdjustR2Dimension;
  protected
    { Protected declarations }
  public
    { Public declarations }
    Constructor Create( aOwner : Tcomponent ); override;
    Destructor Destroy; override;
    Property Canvas;
    Procedure Paint; override;
    //
    Procedure ClrScr;
    Procedure PsetXY( x, y : double; color : Tcolor );
  published
    { Published declarations }
    Property AxesColor : Tcolor read fAxesColor write SetAxesColor;
    Property BackColor : Tcolor read fBackColor write SetBackColor;
    // Distanza dell'area utile per il grafico dai margini del pannello
    // che lo contiene
    Property MarginOriz : integer read fMarginOriz write SetMarginOriz;
    Property MarginVert : integer read fMarginVert write SetMarginVert;
    // Range dei valori numerici visualizzati dal grafico
    Property XMin : double read fXmin write SetXMin;
    Property XMax : double read fXmax write SetXMax;
    Property YMin : double read fYmin write SetYMin;
    Property YMax : double read fYmax write SetYMax;
    // Settaggi per le indicazioni di scala dei valori sugli assi
    Property ScaleX : boolean read fScaleX write SetScaleX;
    Property ScaleXFont : Tfont read fScaleXFont write SetScaleXFont;
    Property ScaleXTags : integer read fScaleXTags write SetScaleXTags;
    Property ScaleY : boolean read fScaleY write SetScaleY;
    Property ScaleYFont : Tfont read fScaleYFont write SetScaleYFont;
    Property ScaleYTags : integer read fScaleYTags write SetScaleYTags;
    //
    Property GraphType : TxgrType read fGraphType write SetGraphType;
    Property PlotMode : TxgrPlotMode read fPlotMode write SetPlotMode;
  end;

procedure Register;

implementation


//
Procedure TmathPanel.SetBackColor( value : Tcolor );
begin
     fBackColor := value;
     //
     Paint;
end;


//
Procedure TmathPanel.SetAxesColor( value : Tcolor );
begin
     fAxesColor := value;
     //
     Paint;
end;


//
Procedure TmathPanel.SetMarginOriz( value : integer );
begin
     fMarginOriz := value;
     Paint;
end;


//
Procedure TmathPanel.SetMarginVert( value : integer );
begin
     fMarginVert := value;
     Paint;
end;


//
Procedure TmathPanel.SetXMin( value : double );
begin
     fXMin := value;
     //                
     Paint;
end;


//
Procedure TmathPanel.SetXMax( value : double );
begin
     fXMax := value;
     //
     Paint;
end;


//
Procedure TmathPanel.SetYMin( value : double );
begin
     fYMin := value;
     //
     Paint;
end;


//
Procedure TmathPanel.SetYMax( value : double );
begin
     fYMax := value;
     //
     Paint;
end;


//
Procedure TmathPanel.SetScaleX( value : boolean );
begin
     fScaleX := value;
     If value then
     begin
          MarginOriz := 30;
          MarginVert := 25;
     end
     else
          MarginVert := 5;
end;


//
Procedure TmathPanel.SetScaleXFont( value : Tfont );
begin
     fScaleXFont.Assign( value );
     //
     Paint;
end;


//
Procedure TmathPanel.SetScaleXTags( value : integer );
begin
end;


//
Procedure TmathPanel.SetScaleY( value : boolean );
begin
     fScaleY := value;
     If value then
     begin
          MarginOriz := 30;
          MarginVert := 25;
     end
     else
          MarginOriz := 5;
end;


//
Procedure TmathPanel.SetScaleYFont( value : Tfont );
begin
     fScaleYFont.Assign( value );
     //
     Paint;
end;


//
Procedure TmathPanel.SetScaleYTags( value : integer );
begin
end;


//
Procedure TmathPanel.SetGraphType( value : TxgrType );
begin
     fGraphType := value;
     //
     Paint;
end;


//
Procedure TmathPanel.SetPlotMode( value : TxgrPlotMode );
begin
     fPlotMode := value;
     Paint;
end;


//
Procedure TmathPanel.AdjustR2Dimension;
begin
     // Width and Height (pixels) of the cartesian area
     Case fGraphType of
     xgrCentered: begin
                       R2Width  := width - ( fMarginOriz * 2 );
                       R2Height := height - ( fMarginVert * 2 );
                  end;
     xgr1Quadrant: begin
                       R2Width  := width - ( fMarginOriz * 2 ) - 5;
                       R2Height := height - ( fMarginVert * 2 ) - 5;
                   end;
     xgrFlatX: begin
                    R2Width  := width - ( fMarginOriz * 2 ) - 5;
                    R2Height := height - ( fMarginVert * 2 ) - 20;
               end;
     xgrFlatY: begin
                    R2Width  := width - ( fMarginOriz * 2 ) - 25;
                    R2Height := height - ( fMarginVert * 2 ) - 5;
               end;
     end;
     // Pixels per numeric unit
     Xqnt := R2Width / (fXMax - fXMin);
     Yqnt := R2Height / (fYMax - fYMin);
end;


// 
Constructor TmathPanel.Create;
begin
     inherited Create( aOwner );
     // Last drawn point coordinates (used by xgrDotModeFast PLOTMODE)
     LastXpixDrawn := -1;
     LastYpixDrawn := -1;
     // Global panel default settings
     Height := 200;
     Width  := 250;
     Caption := '';
     // Graph-backgound and axis colors
     fBackColor := clNavy;
     fAxesColor := clYellow;
     // Default distances between graph-edges and panel borders  
     fMarginOriz := 5;
     fMarginVert := 5;
     // Numerical starting ranges   
     fXMin := -10;
     fXMax := 10;
     fYMin := -10;
     fYMax := 10;
     // Values referred to the axis aspect   
     fScaleX := false;
     fScaleXFont := Tfont.Create;
     fScaleXTags := 3;
     fScaleY := false;
     fScaleYFont := Tfont.Create;
     fScaleYTags := 3;
     //
     fGraphType := xgrCentered;
     fPlotMode  := xgrDotMode;
end;


// 
Destructor TmathPanel.Destroy;
begin
     If assigned( fScaleXFont ) then
     begin
          fScaleXFont.Free;
          fScaleXFont := nil;
     end;
     If assigned( fScaleYFont ) then
     begin
          fScaleYFont.Free;
          fScaleYFont := nil;
     end;
     //
     inherited Destroy;
end;


// 
Procedure TmathPanel.Paint;
Var re : Trect;
    XminStr, XmedStr, XmaxStr,
    YminStr, YmedStr, YmaxStr : string;
begin
     //
     inherited;
     //
     LastXpixDrawn := -1;
     LastYpixDrawn := -1;
     // Refresh proportional value
     AdjustR2Dimension;
     // Draws the window for the graph
     re.Left := fMarginOriz;
     re.Top := fMarginVert;
     re.Right := Width - fMarginOriz;
     re.Bottom := Height - fMarginVert;
     Canvas.Brush.Color := fBackColor;
     Canvas.FillRect( re );
     // Draws cartesian axis
     Case fGraphType of
     xgrCentered: begin
                       Canvas.Pen.Color := fAxesColor;
                       // Y-axis line
                       Canvas.moveTo( width div 2, fMarginVert + 5 );
                       Canvas.lineto( width div 2, height - ( fMarginVert + 5 ) );
                       // X-axis line
                       Canvas.moveTo( fMarginOriz + 5, height div 2 );
                       Canvas.lineto( width - ( fMarginOriz + 5 ), height div 2 );
                  end;
     xgr1quadrant: begin
                       Canvas.Pen.Color := fAxesColor;
                       // Y-axis line
                       Canvas.moveTo( fMarginOriz + 5, fMarginVert + 5 );
                       Canvas.lineto( fMarginOriz + 5, height - ( fMarginVert + 5 ) );
                       // X-axis line
                       Canvas.moveTo( fMarginOriz + 5, height - ( fMarginVert + 5 ) );
                       Canvas.lineto( width - ( fMarginOriz + 5 ), height - ( fMarginVert + 5 ) );
                  end;
     xgrFlatX: begin
                       Canvas.Pen.Color := fAxesColor;
                       // X-axis line
                       Canvas.moveTo( fMarginOriz + 5, height - ( fMarginVert + 20 ) );
                       Canvas.lineto( width - ( fMarginOriz + 5 ), height - ( fMarginVert + 20 ) );
               end;
     xgrFlatY: begin
                       Canvas.Pen.Color := fAxesColor;
                       // Y-axis line
                       Canvas.moveTo( fMarginOriz + 25, fMarginVert + 5 );
                       Canvas.lineto( fMarginOriz + 25, height - ( fMarginVert + 5 ) );
               end;
     end; // CASE
     //
     If fScaleX then
     begin

          // ***********************
          // WRITES X AXIS VALUES

          // Sets X-axis color
          Canvas.Brush.Color := Color;
          Canvas.Font.Assign( fScaleXFont );
          // Writes the l-bound value
          XminStr := FloatToStr( fXMin );
          Canvas.TextOut( fMarginOriz,
                          height - MarginVert + 1,
                          XminStr );
          // Writes the middle value
          If fGraphType <> xgr1Quadrant then
          begin
               XmedStr := FloatToStr( ( fXMax + fXMin ) / 2 );
               Canvas.TextOut( width div 2 - Canvas.TextWidth( XmedStr ) div 2,
                               height - MarginVert + 1,
                               XmedStr );
          end;
          // Writes the u-bound value
          XmaxStr := FloatToStr( fXMax );
          Canvas.TextOut( width - fMarginOriz - Canvas.TextWidth( XmaxStr ),
                          height - MarginVert + 1,
                          XmaxStr );
     end;
     // 
     If fScaleY then
     begin
          // ***********************
          // WRITES X AXIS VALUES

          // Sets Y-axis color
          Canvas.Brush.Color := Color;
          Canvas.Font.Assign( fScaleYFont );
          // Writes the l-bound value
          YmaxStr := FloatToStr( fYMax );
          Canvas.TextOut( width - fMarginOriz + 3,
                          fMarginVert,
                          YmaxStr );
          // Writes the middle value
          If fGraphType <> xgr1Quadrant then
          begin
               YmedStr := FloatToStr( ( fYMax + fYMin ) / 2 );
               Canvas.TextOut( width - fMarginOriz + 3,
                               height div 2 - Canvas.TextWidth( YmedStr ) div 2,
                               YmedStr );
          end;
          // Writes the u-bound value
          YminStr := FloatToStr( fYMin );
          Canvas.TextOut( width - fMarginOriz + 3,
                          height - fMarginVert - Canvas.TextHeight( YminStr ),
                          YminStr );

     end;
end;


// Clears the graph-area
Procedure TmathPanel.ClrScr;
begin
     Paint;
end;


// 
Procedure TmathPanel.PsetXY( x, y : double; color : Tcolor );
begin
     if (y<fYMin) or
        (y>fYMax) then
        Exit;
     if (x<fXMin) or
        (x>fXMax) then
        Exit;
     //
     Case fGraphType of
     xgrCentered:
     begin
          Case fPlotMode of
          xgrDotMode: Canvas.Pixels[fMarginOriz +
                                    trunc(abs(x - fXMin)* Xqnt) ,
                                    height - fMarginVert -
                                    trunc(abs(y - fYMin)* Yqnt)] := color;
          xgrDotModeFast: begin
                               If ( LastXpixDrawn <> -1 ) and
                                  ( LastYpixDrawn <> -1 ) then
                               begin
                                    Canvas.moveTo( LastXpixDrawn,
                                                   LastYpixDrawn );
                                    Canvas.LineTo( fMarginOriz +
                                                   trunc(abs(x - fXMin)* Xqnt) ,
                                                   height - fMarginVert -
                                                   trunc(abs(y - fYMin)* Yqnt) );
                                    LastXpixDrawn := fMarginOriz +
                                                     trunc(abs(x - fXMin)* Xqnt);
                                    LastYpixDrawn := height - fMarginVert -
                                                     trunc(abs(y - fYMin)* Yqnt);
                               end
                               else
                               begin
                                    Canvas.Pixels[fMarginOriz +
                                                  trunc(abs(x - fXMin)* Xqnt) ,
                                                  height - fMarginVert -
                                                  trunc(abs(y - fYMin)* Yqnt)] := color;
                                    LastXpixDrawn := fMarginOriz +
                                                     trunc(abs(x - fXMin)* Xqnt);
                                    LastYpixDrawn := height - fMarginVert -
                                                     trunc(abs(y - fYMin)* Yqnt);
                               end;
                          end;
          xgrBarMode: begin
                           Canvas.moveTo(fMarginOriz +
                                         trunc(abs(x - fXMin)* Xqnt) ,
                                         height - fMarginVert -
                                         trunc(abs(y - fYMin)* Yqnt));
                           Canvas.lineTo(fMarginOriz +
                                         trunc(abs(x - fXMin)* Xqnt),
                                         height div 2);
                      end;
          end; // case
     end;
     xgr1Quadrant:
     begin
          Case fPlotMode of
          xgrDotMode: Canvas.Pixels[fMarginOriz + 5 +
                                    trunc(abs(x - fXMin)* Xqnt) ,
                                    height - fMarginVert - 5 -
                                    trunc(abs(y - fYMin)* Yqnt)] := color;
          xgrDotModeFast: begin
                               If ( LastXpixDrawn <> -1 ) and
                                  ( LastYpixDrawn <> -1 ) then
                               begin
                                    Canvas.moveTo( LastXpixDrawn,
                                                   LastYpixDrawn );
                                    Canvas.LineTo( fMarginOriz + 5 +
                                                   trunc(abs(x - fXMin)* Xqnt) ,
                                                   height - fMarginVert - 5 -
                                                   trunc(abs(y - fYMin)* Yqnt) );
                                    LastXpixDrawn := fMarginOriz + 5 +
                                                     trunc(abs(x - fXMin)* Xqnt);
                                    LastYpixDrawn := height - fMarginVert - 5 -
                                                     trunc(abs(y - fYMin)* Yqnt);
                               end
                               else
                               begin
                                    Canvas.Pixels[fMarginOriz + 5 +
                                                  trunc(abs(x - fXMin)* Xqnt) ,
                                                  height - fMarginVert - 5 -
                                                  trunc(abs(y - fYMin)* Yqnt)] := color;
                                    LastXpixDrawn := fMarginOriz + 5 +
                                                     trunc(abs(x - fXMin)* Xqnt);
                                    LastYpixDrawn := height - fMarginVert - 5 -
                                                     trunc(abs(y - fYMin)* Yqnt);
                               end;
                          end;
          xgrBarMode: begin
                           Canvas.MoveTo(fMarginOriz + 5 +
                                         trunc(abs(x - fXMin)* Xqnt) ,
                                         height - fMarginVert - 5 -
                                         trunc(abs(y - fYMin)* Yqnt) );
                           Canvas.lineTo(fMarginOriz + 5 +
                                         trunc(abs(x - fXMin)* Xqnt) ,
                                         height - fMarginVert - 5);
                      end;
          end; // case
     end;
     xgrFlatX:
     begin
          Case fPlotMode of
          xgrDotMode: Canvas.Pixels[fMarginOriz + 5 +
                                    trunc(abs(x - fXMin)* Xqnt) ,
                                    height - fMarginVert - 20 -
                                    trunc(abs(y - fYMin)* Yqnt)] := color;
          xgrDotModeFast: begin
                               If ( LastXpixDrawn <> -1 ) and
                                  ( LastYpixDrawn <> -1 ) then
                               begin
                                    Canvas.moveTo( LastXpixDrawn,
                                                   LastYpixDrawn );
                                    Canvas.LineTo( fMarginOriz + 5 +
                                                   trunc(abs(x - fXMin)* Xqnt) ,
                                                   height - fMarginVert - 20 -
                                                   trunc(abs(y - fYMin)* Yqnt) );
                                    LastXpixDrawn := fMarginOriz + 5 +
                                                     trunc(abs(x - fXMin)* Xqnt);
                                    LastYpixDrawn := height - fMarginVert - 20 -
                                                     trunc(abs(y - fYMin)* Yqnt);
                               end
                               else
                               begin
                                    Canvas.Pixels[fMarginOriz + 5 +
                                                  trunc(abs(x - fXMin)* Xqnt) ,
                                                  height - fMarginVert - 20 -
                                                  trunc(abs(y - fYMin)* Yqnt)] := color;
                                    LastXpixDrawn := fMarginOriz + 5 +
                                                     trunc(abs(x - fXMin)* Xqnt);
                                    LastYpixDrawn := height - fMarginVert - 20 -
                                                     trunc(abs(y - fYMin)* Yqnt);
                               end;
                          end;
          xgrBarMode: begin
                           Canvas.MoveTo(fMarginOriz + 5 +
                                         trunc(abs(x - fXMin)* Xqnt) ,
                                         height - fMarginVert - 20 -
                                         trunc(abs(y - fYMin)* Yqnt));
                           Canvas.LineTo(fMarginOriz + 5 +
                                         trunc(abs(x - fXMin)* Xqnt) ,
                                         height - fMarginVert - 20);
                      end;
          end; // case
     end;
     xgrFlatY:
     begin
          Case fPlotMode of
          xgrDotMode: Canvas.Pixels[fMarginOriz + 25 +
                                    trunc(abs(x - fXMin)* Xqnt) ,
                                    height - fMarginVert - 5 -
                                    trunc(abs(y - fYMin)* Yqnt)] := color;
          xgrDotModeFast: begin
                               If ( LastXpixDrawn <> -1 ) and
                                  ( LastYpixDrawn <> -1 ) then
                               begin
                                    Canvas.moveTo( LastXpixDrawn,
                                                   LastYpixDrawn );
                                    Canvas.LineTo( fMarginOriz + 25 +
                                                   trunc(abs(x - fXMin)* Xqnt) ,
                                                   height - fMarginVert - 5 -
                                                   trunc(abs(y - fYMin)* Yqnt) );
                                    LastXpixDrawn := fMarginOriz + 25 +
                                                     trunc(abs(x - fXMin)* Xqnt);
                                    LastYpixDrawn := height - fMarginVert - 5 -
                                                     trunc(abs(y - fYMin)* Yqnt);
                               end
                               else
                               begin
                                    Canvas.Pixels[fMarginOriz + 25 +
                                                  trunc(abs(x - fXMin)* Xqnt) ,
                                                  height - fMarginVert - 5 -
                                                  trunc(abs(y - fYMin)* Yqnt)] := color;
                                    LastXpixDrawn := fMarginOriz + 25 +
                                                     trunc(abs(x - fXMin)* Xqnt);
                                    LastYpixDrawn := height - fMarginVert - 5 -
                                                     trunc(abs(y - fYMin)* Yqnt);
                               end;
                          end;
          xgrBarMode: begin
                           Canvas.MoveTo(fMarginOriz + 25 +
                                         trunc(abs(x - fXMin)* Xqnt) ,
                                         height - fMarginVert - 5 -
                                         trunc(abs(y - fYMin)* Yqnt));
                           Canvas.LineTo(fMarginOriz + 25,
                                         height - fMarginVert - 5 -
                                         trunc(abs(y - fYMin)* Yqnt));
                      end;
          end; // case
     end;
     end; // case
end;


procedure Register;
begin
  RegisterComponents('Additional', [TmathPanel]);
end;


end.

