unit SystemInfo;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Registry;

type
  TSystemInfo = class(TComponent)
  private
    { Private declarations }
    FComputerName: String;
    FUserName: String;    
    FWinDir: String;
    FSysDir: String;
    FRegOwner: String;
    FRegCompany: String;
    FVersion: String;
    FPlatform: String;
    FPlatformID: DWord;
  protected
    { Protected declarations }
    constructor Create(AOwner: TComponent); override;
    procedure GetComputer;
    procedure GetUser;
    procedure GetDirectories;
    procedure GetRegInfo;
    procedure GetOSVersion;
  public
    { Public declarations }
  published
    { Published declarations }
    property ComputerName: String read FComputerName;
    property UserName: String read FUserName;
    property WinDir: String read FWinDir;
    property SysDir: String read FSysDir;
    property RegOwner: String read FRegOwner;
    property RegCompany: String read FRegCompany;
    property OSVersion: String read FVersion;
    property OSPlatform: String read FPlatform;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TSystemInfo]);
end;

constructor TSystemInfo.Create(AOwner: TComponent);
begin
   inherited;
   //get details
   GetComputer;
   GetUser;
   GetDirectories;
   GetOSVersion;   
   GetRegInfo;     //getosversion must be called before getreginfo   
end;

procedure TSystemInfo.GetComputer;
var
   ComputerName : PChar;
   NameSize : DWORD;
begin
   ComputerName := #0;
   NameSize := MAX_COMPUTERNAME_LENGTH + 1;
   try
      GetMem(ComputerName, NameSize);
      Windows.GetComputerName(ComputerName, NameSize);
      FComputerName:= StrPas(ComputerName);
   finally
      FreeMem(ComputerName);
   end;
end;

procedure TSystemInfo.GetUser;
var
   UserName : PChar;
   NameSize : DWORD;
begin
   UserName := #0;
   NameSize := 50;
   try
      GetMem(UserName, NameSize);
      Windows.GetUserName(UserName, NameSize);
      FUserName:= StrPas(UserName);
   finally
      FreeMem(UserName);
   end;
end;

procedure TSystemInfo.GetDirectories;
var 
    Size: integer;
    Names: Pchar;
begin
  Names := #0;
  try
    GetMem( Names, MAX_PATH+1 );

    Size := GetSystemDirectory(Names, MAX_PATH+1);
    FSysDir := StrPas(Names);
    
    Size := GetWindowsDirectory(Names, MAX_PATH+1);
    FWinDir := StrPas(Names);
    
  finally
    FreeMem( Names );
  end; //of try
end; 

procedure TSystemInfo.GetRegInfo;
const
  WIN95_KEY = '\SOFTWARE\Microsoft\Windows\CurrentVersion';
  WINNT_KEY = '\SOFTWARE\Microsoft\Windows NT\CurrentVersion';
var
  VersionKey : PChar;
begin
   //get platform id details  
   //    note platform id is set by getosversion - so getosversion should already have been called
   case FPlatformID of
      VER_PLATFORM_WIN32_WINDOWS:  VersionKey := WIN95_KEY;
      VER_PLATFORM_WIN32_NT     :  VersionKey := WINNT_KEY;
      else  VersionKey := nil;
   end;

   with TRegistry.Create do
   try
       RootKey := HKEY_LOCAL_MACHINE;

       if OpenKey(VersionKey, False) then 
          begin
          FRegOwner := ReadString('RegisteredOwner');
          FRegCompany := ReadString('RegisteredOrganization');
          end; 
   finally
     Free;
   end; // of try
end; 

procedure TSystemInfo.GetOSVersion;
var
  OSVersion: TOSVersionInfo;
begin
   OSVersion.dwOSVersionInfoSize := SizeOf(OSVersion);
      
   if GetVersionEx(OSVersion) then
      begin
         FVersion := Format('%d.%d (%d.%s)',[OSVersion.dwMajorVersion, OSVersion.dwMinorVersion,(OSVersion.dwBuildNumber and $FFFF), OSVersion.szCSDVersion]);
         FPlatformID:= OSVersion.dwPlatformID;
         case OSVersion.dwPlatformID of
            VER_PLATFORM_WIN32s:        FPlatform := 'Windows 3.1';
            VER_PLATFORM_WIN32_WINDOWS: FPlatform := 'Windows 95';
            VER_PLATFORM_WIN32_NT:      FPlatform := 'Windows NT';
            else                        FPlatform := '';
         end;  //of case
      end; //of if
end; 

end.
